local wf = AceLibrary("Waterfall-1.0");
local pc = AceLibrary("PaintChips-2.0");
local L = AceLibrary("AceLocale-2.2"):new("SorrenTimers");
local sm = AceLibrary("SharedMedia-1.0");
local tablet = AceLibrary("Tablet-2.0");

--bummed from ckknight's pitbull, with his permission:
local new, del, newHash
do
	local list = setmetatable({}, {__mode='k'})
	function new(...)
		local t = next(list)
		if t then
			list[t] = nil
			for i = 1, select('#', ...) do
				t[i] = select(i, ...)
			end
			return t
		else
			return { ... }
		end
	end
	--[[ don't think I particularly need these
	function newHash(...)
		local t = next(list)
		if t then
			list[t] = nil
		else
			t = {}
		end
		for i = 1, select('#', ...), 2 do
			t[select(i, ...)] = select(i+1, ...)
		end
		return t
	end

	function newSet(...)
		local t = next(list)
		if t then
			list[t] = nil
		else
			t = {}
		end
		for i = 1, select('#', ...) do
			t[select(i, ...)] = true
		end
		return t
	end
	--]]
	function del(t)
		for k in pairs(t) do
			t[k] = nil
		end
		list[t] = true
		return nil
	end
end
--end sleezy code borrowing

local db;

SorrenTimers.listGroups = {};
SorrenTimers.skillOptions = {
	defaultgroup = {
		type = 'text',
		name = L["Default Group"],
		desc = L["Set the default Group for these bars"],
		order = 1,
		get = function()
			return SorrenTimers.db.char.defaultskillgroup;
		end,
		set = function(g)
			SorrenTimers.db.char.defaultskillgroup = g;
		end,
		validate = SorrenTimers.listGroups,
	},
};
SorrenTimers.enemySkillOptions = {
	defaultgroup = {
		type = 'text',
		name = L["Default Group"],
		desc = L["Set the default Group for these bars"],
		order = 1,
		get = function()
			return SorrenTimers.db.char.defaultenemyskillgroup;
		end,
		set = function(g)
			SorrenTimers.db.char.defaultenemyskillgroup = g;
		end,
		validate = SorrenTimers.listGroups,
	},

};
local scolors = {};
SorrenTimers.colorOptions = {
  newcolor = {
    order = 1,
    type = 'text',
    name = L["New Custom Color"],
    desc = L["Creates a new custom color."],
    get = function()
      return "";
    end,
    set = function(v)
      if v then v = v:lower(); end
      if pc.vars.hex[v] or SorrenTimers.db.char.customcolors[v] then
        SorrenTimers:Print(string.format(L["The color %s is already defined."], v));
      elseif not v then
        SorrenTimers:Print(L["Error: Field empty"]);
      else
        SorrenTimers.db.char.customcolors[v] = new(1,1,1);
        pc:RegisterColor(v,1,1,1);
        SorrenTimers:SetColors();
        wf:Refresh("SorrenTimers");
      end
    end,
  },
  delcolor = {
    type = 'text',
    order = 2,
    name = L["Delete Color"],
    desc = L["Remove a custom color"],
    get = function()
      return "";
    end,
    set = function(v)
      pc:UnregisterColor(v);
      SorrenTimers.db.char.customcolors[v] = nil;
      SorrenTimers.colorOptions[v] = nil;
      SorrenTimers:SetColors();
      wf:Refresh("SorrenTimers");
    end,
    validate = scolors,
  },
};

SorrenTimers.groupSkillOptions = {
	defaultgroup = {
		type = 'text',
		name = L["Default Group"],
		desc = L["Set the default Group for these bars"],
		order = 1,
		get = function()
			return SorrenTimers.db.char.defaultgroupskillsgroup;
		end,
		set = function(g)
			SorrenTimers.db.char.defaultgroupskillsgroup = g;
		end,
		validate = SorrenTimers.listGroups,
	},
  enabled = {
    type = 'toggle',
    name = L["Enabled"],
    desc = L["Enable receiving spell information from other players."],
    order = 0,
    get = function()
      return not SorrenTimers.db.char.groups.groupskills.disabled;
    end,
    set = function(v)
      SorrenTimers.db.char.groups.groupskills.disabled = not v;
      if v then
        if SorrenTimers.db.char.groups.groupskills.locked then
          SorrenTimers.anchors.groupskills:Hide();
        else
          SorrenTimers.anchors.groupskills:Show();
        end
        if GetNumPartyMembers() > 0 then
          SorrenTimers:RequestOffsets();
          SorrenTimers:Party();
        end
        SorrenTimers:RegisterComm("SorrenTimers", "GROUP");
      else
        SorrenTimers.anchors.groupskills:Hide();
        SorrenTimers:UnregisterComm("SorrenTimers");
      end
    end,
  },

};
SorrenTimers.buffOptions = {
	defaultgroup = {
		type = 'text',
		name = L["Default Group"],
		desc = L["Set the default Group for these bars"],
		order = 1,
		get = function()
			return SorrenTimers.db.char.defaultbuffgroup;
		end,
		set = function(g)
			SorrenTimers.db.char.defaultbuffgroup = g;
		end,
		validate = SorrenTimers.listGroups,
	},
};

SorrenTimers.groupOptions = {
	createnew = {
		type = 'text',
		name = L["Create New Group"],
		desc = L["Create a new bar grouping"],
		order = 1,
		get = function() return ""; end,
		set = function(v) 
			if string.find(v, "[^a-zA-Z]") then
				SorrenTimers:Print(L["Group names need to be letters only."]);
				return;
			elseif SorrenTimers.db.char.groups[v] then
				SorrenTimers:Print(string.format(L["Group %s already exists."], v));
				return;
			else
				SorrenTimers:CreateNewGroup(v:lower());
			end
		end,
	},
};
	
SorrenTimers.wfoptions = {
	type = "group",
	args = {
		defaultoptions = {
			type = "group",
			name = L["Defaults"],
			desc = L["Set the default look and feel of the bars"],
			order = 2,
			args = {
			
				colors = {
					type = "text",
					name = L["Bar Colors"],
					desc = L["Sets the colors for the bars"],
					usage = L["<color1> <color2> <color3>...(must have at at least 1 color)"],
					get = function() 
						local c = "";
						for k,v in pairs(SorrenTimers.db.char.defaults.colors) do
							c = c..v.." ";
						end
						return c;
					end,
					set = function(s) 
						local c = new();
						for v in string.gmatch(s, "(%w+)") do
							if pc:GetHex(v) then
								table.insert(c, v);
							else
								SorrenTimers:Print(string.format(L["Color %s is not a valid color, skipping."], v));
							end
						end
            if next(c) then
  						del(SorrenTimers.db.char.defaults.colors);
  						SorrenTimers.db.char.defaults.colors = c;
            else	
              SorrenTimers:Print(L["No colors were valid."]);
            end
					end,
				},
				dimincolors = {
					type = "text",
					name = L["Diminishing Returns Bar Colors"],
					desc = L["Sets the colors for the bars"],
					usage = L["<color1> <color2> <color3>...(must have at at least 1 color)"],
					get = function() 
						local c = "";
						for k,v in pairs(SorrenTimers.db.char.defaults.diminishcolors) do
							c = c..v.." ";
						end
						return c;
					end,
					set = function(s) 
						local c = new();
						for v in string.gmatch(s, "(%w+)") do
							if pc:GetHex(v) then
								table.insert(c, v);
							else
								SorrenTimers:Print(string.format(L["Color %s is not a valid color, skipping."], v));
							end
						end
            if next(c) then
  						del(SorrenTimers.db.char.defaults.diminishcolors);
  						SorrenTimers.db.char.defaults.diminishcolors = c;
            else	
              SorrenTimers:Print(L["No colors were valid."]);
            end
					end,
				},
				fade = {
					type = "text",
					name = L["Fade Time"],
					desc = L["How long it takes the bar to fade after completion"],
					usage = L["<time> (in seconds, use a negative time for it to stick and never fade)"],
					get = function()
						return SorrenTimers.db.char.defaults.fade;
					end,
					set = function(s)
						local n = tonumber(s);
						if not n then
							SorrenTimers:Print(string.format(L["%s is not a valid number."], s));
						else
							SorrenTimers.db.char.defaults.fade = n;
						end
					end,
				},
				width = {
					type = "range",
					name = L["Bar Width"],
					desc = L["How wide bars are"],
					min = 32,
					max = 200,
					step = 1,
					get = function()
						return SorrenTimers.db.char.defaults.width;
					end,
					set = function(s)
						SorrenTimers.db.char.defaults.width = s;
					end,
				},
				height = {
					type = "range",
					name = L["Bar Height"],
					desc = L["How tall bars are"],
					min = 16,
					max = 32,
					step = 1,
					get = function()
						return SorrenTimers.db.char.defaults.height;
					end,
					set = function(s)
						SorrenTimers.db.char.defaults.height = s;
					end,
				},
				growth = {
					type = "toggle",
					name = L["Grow Upwards"],
					desc = L["Controls the direction in which the bars of this group will grow"],
					get = function()
						return SorrenTimers.db.char.defaults.up;
					end,
					set = function(v)
						SorrenTimers.db.char.defaults.up = v;
						for name,data in pairs(SorrenTimers.db.char.groups) do
							if data.up ~= nil then
								SorrenTimers:SetCandyBarGroupGrowth(name, data.up)
							else
								SorrenTimers:SetCandyBarGroupGrowth(name, v)
							end
						end
					end,
					--map = { [true] = L["Upwards"], [false] = L["Downwards"] },
				},
				targetstring = {
					type = "text",
					name = L["Display String (target)"],
					desc = L["How bars that have a target will display\nUse $name for skill name, $target for skill target, \nex. $name - $target"],
					usage = L["use $name for where the skill goes, $target for the target (ex. $name - $target)"],
					get = function()
						return SorrenTimers.db.char.defaults.targetstring;
					end,
					set = function(s)
						SorrenTimers.db.char.defaults.targetstring = s;
					end,
				},
				skillstring = {
					type = "text",
					name = L["Display String (no target)"],
					desc = L["How bars will display the skill name, use $name for skill name"],
					usage = L["use $name for where the skill goes"],
					get = function()
						return SorrenTimers.db.char.defaults.skillstring;
					end,
					set = function(s)
						SorrenTimers.db.char.defaults.skillstring = s;
					end,
				},
				diminishstring = {
					type = "text",
					name = L["Diminish String (target)"],
					desc = L["How bars that have a target will display, use $n for number of applications\n$name for skill name, $target for target,\nex. ($n) $name - $target"],
					usage = L["use $name for where the skill goes, $n for the applications, $target for target"],
					get = function()
						return SorrenTimers.db.char.defaults.diminishstring;
					end,
					set = function(s)
						SorrenTimers.db.char.defaults.diminishstring = s;
					end,
				},
				texture = {
					type = 'text',
					name = L["Bar Texture"],
					desc = L["The texture bars will use"],
					usage = L["<texture>"],
					get = function()
						return SorrenTimers.db.char.defaults.texture;
					end,
					set = function(s)
						SorrenTimers.db.char.defaults.texture = s;
					end,
          columns = 2,
					validate = sm:List("statusbar"),
				},
        scale = {
          type = 'range',
          name = L["Scale"],
          desc = L["The Scale of the group's bars"],
          min = 0.25,
          max = 2,
          step = 0.05,
          get = function()
            return SorrenTimers.db.char.defaults.scale;
          end,
          set = function(v)
            SorrenTimers.db.char.defaults.scale = v;
          end,
        },
			},
		},
		groupoptions = {
			type = "group",
			name = L["Group Options"],
			desc = L["Set the look and feel for the bars of a partciluar group"],
			order = 2,
			args = SorrenTimers.groupOptions,
		},
		skilloptions = {
			type = 'group',
			name = L["Skill Options"],
			desc = L["Change the look and feel of individual bars. This section is populated as you use more skills."],
			order = 3,
			args = SorrenTimers.skillOptions,
		},
		buffoptions = {
			type = 'group',
			name = L["Buff Tracking"],
			desc = L["Track certain buffs with status bars, must be individually enabled. This is populated when you gain trackable buffs."],
			order = 4,
			args = SorrenTimers.buffOptions,
		},
		enemyskilloptions = {
			type = 'group',
			name = L["Enemy Skill Options"],
			desc = L["Change the look and feel of individual bars pertaining to enemy abilities"],
			order = 5,
			args = SorrenTimers.enemySkillOptions,
		},
    groupskilloptions = {
      type = 'group',
      name = L["Group Skill Options"],
      desc = L["Enable/disable/customize information you receive from other players. This list is populated when you receive new ability information."],
      order = 6,
      args = SorrenTimers.groupSkillOptions,
    },
    coloroptions = {
      type = 'group',
      name = L["Color Options"],
      desc = L["Make new colors to use in the bars. Default paintchip colors cannot be changed."],
      order = 7,
      args = SorrenTimers.colorOptions,
    },
	},
};

function SorrenTimers:CreateNewGroup(v)
	db.groups[v] = {};
	self:CreateAnchor(v);
	self:RegisterCandyBarGroup(v);
	self:SetCandyBarGroupPoint(v, "TOP", self.anchors[v]:GetName(), "BOTTOM", 0, 0)
	self:SetCandyBarGroupGrowth(v, db.defaults.up);
	self:SetGroups();
end

function SorrenTimers:DeleteGroup(v)
	db.groups[v] = nil;
	self.anchors[v]:Hide();
	
	for k,v in pairs(db.custom) do
		if v.group == v then
			db.custom[k].group = nil;
		end
	end
	self.groupOptions[v] = nil;
	self:SetGroups();
end

function SorrenTimers:SetGroups()
	del(self.listGroups);
	self.listGroups = new();
	for k in pairs(db.groups) do
		if not self.groupOptions[k] then
			self.groupOptions[k] = {
				type = "group",
				name = k,
				args = SorrenTimers:GetGroupOptions(k),
			};
      if k == "groupskills" then
        self.groupOptions[k].disabled = function() return self.db.char.groups[k].disabled end;
      end
		end
    if k == "groupskills" then
      self.groupOptions[k].disabled = function() return self.db.char.groups[k].disabled end;
    end
		table.insert(self.listGroups, k);
	end
	table.sort(self.listGroups);
end

function SorrenTimers:OnClick()
	self:OpenOptions();
end

function SorrenTimers:GetGroupOptions(k)
	local v = {
		locked = {
			type = 'toggle',
			name = L["Locked"],
			desc = L["Show/hide this group's anchor"],
			order = 1,
			get = function()
				return db.groups[k].locked;
			end,
			set = function(v)
				if not v then
					self.anchors[k]:Show();
				else
					self.anchors[k]:Hide();
				end
				db.groups[k].locked = v;
			end,
		},
		resetpos = {
			type = 'execute',
			name = L["Reset Location"],
			desc = L["Use this to bring the skill back to the center of the screen if you've accidentally dragged it off."],
			order = 18,
			func = function()
				self.anchors[k]:ClearAllPoints();
				self.anchors[k]:SetPoint("CENTER",UIParent,"CENTER",0,0);
			end,
		},
		delete = {
			type = 'execute',
			name = L["Delete Group"],
			desc = L["Delete this group (defaultgroup and BGTimers cannot be deleted)"],
			disabled = function() return k == "defaultgroup" or k == "groupskills" end,
			order = 19,
			func = function()
				SorrenTimers:DeleteGroup(k);
			end,
			confirm = string.format(L["Are you sure you want to delete group %s? All settings for this group will be lost."], k),
		},
		colordefaults = {
			type = "toggle",
			name = L["Use Default Colors"],
			desc = L["Use default settings"],
			order = 20,
			get = function()
				return db.groups[k].colors == nil;
			end,
			set = function(v)
				if not v then
					db.groups[k].colors = db.defaults.colors;
				else
					db.groups[k].colors = nil;
				end
			end,
		},
		colors = {
			type = "text",
			name = L["Bar Colors"],
			desc = L["Sets the colors for the bars"],
			usage = L["<color1> <color2> <color3>...(must have at at least 1 color)"],
			disabled = function() return db.groups[k].colors == nil; end,
			order = 21,
			get = function() 
				local c = "";
				for k,v in pairs(db.groups[k].colors or db.defaults.colors) do
					c = c..v.." ";
				end
				return c;
			end,
			set = function(s) 
				local c = new();
				for v in string.gmatch(s, "(%w+)") do
					if pc:GetHex(v:lower()) then
						table.insert(c, v:lower());
					else
						SorrenTimers:Print(string.format(L["Color %s is not a valid color, skipping."], v));
					end
				end
        if next(c) then
          if db.groups[k].colors then del(db.groups[k].colors); end
          db.groups[k].colors = c;
				else	
					SorrenTimers:Print(L["No colors were valid."]);
        end
			end,
		},
		fadedefaults = {
			type = "toggle",
			name = L["Use Default Fade Time"],
			desc = L["Use default settings"],
			order = 22,
			get = function()
				return db.groups[k].fade == nil;
			end,
			set = function(v)
				if not v then
					db.groups[k].fade = db.defaults.fade;
				else
					db.groups[k].fade = nil;
				end
			end,
		},
		fade = {
			type = "text",
			name = L["Fade Time"],
			desc = L["How long it takes the bar to fade after completion"],
			usage = L["<time> (in seconds, use a negative time for it to stick and never fade)"],
			order = 23,
			disabled = function()
				return db.groups[k].fade == nil;
			end,
			get = function()
				return db.groups[k].fade or db.defaults.fade;
			end,
			set = function(s)
				local n = tonumber(s);
				if not n then
					SorrenTimers:Print(string.format(L["%s is not a valid number."], s));
				else
					db.groups[k].fade = n;
				end
			end,
		},
		widthdefaults = {
			type = "toggle",
			name = L["Use Default Width"],
			desc = L["Use default settings"],
			order = 25,
			get = function()
				return db.groups[k].width == nil;
			end,
			set = function(v)
				if not v then
					db.groups[k].width = db.defaults.width;
				else
					db.groups[k].width = nil;
				end
			end,
		},
		width = {
			type = "range",
			name = L["Bar Width"],
			desc = L["How wide bars are"],
			min = 32,
			max = 200,
			step = 1,
			order = 26,
			disabled = function() return db.groups[k].width == nil end,
			get = function()
				return db.groups[k].width or db.defaults.width;
			end,
			set = function(s)
				db.groups[k].width = s;
			end,
		},
		heightdefaults = {
			type = "toggle",
			name = L["Use Default Height"],
			desc = L["Use default settings"],
			order = 27,
			get = function()
				return db.groups[k].height == nil;
			end,
			set = function(v)
				if not v then
					db.groups[k].height = db.defaults.height;
				else
					db.groups[k].height = nil;
				end
			end,
		},
		height = {
			type = "range",
			name = L["Bar Height"],
			desc = L["How tall bars are"],
			min = 16,
			max = 32,
			step = 1,
			order = 28,
			disabled = function() return db.groups[k].height == nil end,
			get = function()
				return db.groups[k].height or db.defaults.height;
			end,
			set = function(s)
				db.groups[k].height = s;
			end,
		},
		growdefaults = {
			type = "toggle",
			name = L["Use Default Growth Setting"],
			desc = L["Use default settings"],
			order = 29,
			get = function()
				return db.groups[k].up == nil;
			end,
			set = function(v)
				if not v then
					db.groups[k].up = db.defaults.up;
				else
					db.groups[k].up = nil;
				end
				self:SetCandyBarGroupGrowth(k, db.defaults.up)
			end,
		},
		growth = {
			type = "toggle",
			name = L["Grow Upwards"],
			desc = L["Controls the direction in which the bars of this group will grow"],
			order = 30,
			disabled = function() return db.groups[k].up == nil end,
			get = function()
				return db.groups[k].up;
			end,
			set = function(v)
				db.groups[k].up = v;
				self:SetCandyBarGroupGrowth(k, v);
			end,
			--map = { [true] = L["Upwards"], [false] = L["Downwards"] },
		},
		targetstringdefaults = {
			type = "toggle",
			name = L["Use Default Display (target) String"],
			desc = L["Use default settings"],
			order = 31,
			get = function()
				return db.groups[k].targetstring == nil;
			end,
			set = function(v)
				if not v then
					db.groups[k].targetstring = db.defaults.targetstring;
				else
					db.groups[k].targetstring = nil;
				end
			end,
		},
		targetstring = {
			type = "text",
			name = L["Display String (target)"],
			desc = L["How bars that have a target will display\nUse $name for skill name, $target for skill target, \nex. $name - $target"],
			order = 32,
			disabled = function() return db.groups[k].targetstring == nil end,
			get = function()
				return db.groups[k].targetstring or db.defaults.targetstring;
			end,
			set = function(s)
				db.groups[k].targetstring = s;
			end,
		},
		skillstringdefaults = {
			type = "toggle",
			name = L["Use Default Display String"],
			desc = L["Use default settings"],
			order = 32,
			get = function()
				return db.groups[k].skillstring == nil;
			end,
			set = function(v)
				if not v then
					db.groups[k].skillstring = db.defaults.skillstring;
				else
					db.groups[k].skillstring = nil;
				end
			end,
		},
		skillstring = {
			type = "text",
			name = L["Display String (no target)"],
			desc = L["How bars will display the skill name, use $name for skill name"],
			order = 34,
			disabled = function() return db.groups[k].skillstring == nil end,
			get = function()
				return db.groups[k].skillstring or db.defaults.skillstring;
			end,
			set = function(s)
				db.groups[k].skillstring = s;
			end,
		},
		diminishstringdefaults = {
			type = "toggle",
			name = L["Use Default Diminish String"],
			desc = L["Use default settings"],
			order = 35,
			get = function()
				return db.groups[k].diminishstring == nil;
			end,
			set = function(v)
				if not v then
					db.groups[k].diminishstring = db.defaults.diminishstring;
				else
					db.groups[k].diminishstring = nil;
				end
			end,
		},
		diminishstring = {
			type = "text",
			name = L["Diminish String (target)"],
			desc = L["How bars that have a target will display, use $n for number of applications\n$name for skill name, $target for target,\nex. ($n) $name - $target"],
			order = 36,
			disabled = function() return db.groups[k].diminishstring == nil end;
			get = function()
				return db.groups[k].diminishstring or db.defaults.diminishstring;
			end,
			set = function(s)
				db.groups[k].diminishstring = s;
			end,
		},
		texturedefaults = {
			type = "toggle",
			name = L["Use Default Texture"],
			desc = L["Use default settings"],
			order = 37,
			get = function()
				return db.groups[k].texture == nil;
			end,
			set = function(v)
				if not v then
					db.groups[k].texture = db.defaults.texture;
				else
					db.groups[k].texture = nil;
				end
			end,
		},
		texture = {
			type = 'text',
			name = L["Bar Texture"],
			desc = L["The texture bars will use"],
			order = 38,
			disabled = function() return db.groups[k].texture == nil end,
			get = function()
				return db.groups[k].texture or db.defaults.texture;
			end,
			set = function(s)
				db.groups[k].texture = s;
			end,
			validate = sm:List("statusbar"),
		},
    defaultscale = {
      type = 'toggle',
      name = L["Use Default Scale Settings"],
      desc = L["Use Default Scale Settings"],
      order = 39,
      get = function()
        return db.groups[k].scale == nil;
      end,
      set = function(v)
        if not v then
          db.groups[k].scale = db.defaults.scale;
        else
          db.groups[k].scale = nil;
        end
      end,
    },
    scale = {
      type = 'range',
      name = L["Scale"],
      desc = L["The Scale of the group's bars"],
      min = 0.25,
      max = 2,
      step = 0.05,
      disabled = function()
        return db.groups[k].scale == nil;
      end,
      get = function()
        return db.groups[k].scale or db.defaults.scale;
      end,
      set = function(v)
        db.groups[k].scale = v;
      end,
    },
	};
	return v;
end

function SorrenTimers:SetSkills()
	local st = {};
	for k,v in pairs(db.custom) do
		table.insert(st,k);
	end
	table.sort(st);
	for k,v in pairs(st) do
		if not self.skillOptions[v] then
			self.skillOptions[v] = {
				type = 'group',
				name = v,
				order = k,
				args = self:GetSkillOptions(v);
			};
		else
			self.skillOptions[v].order = k;
		end
	end
end


function SorrenTimers:SetGroupSkills()
	local st = {};
	for k,v in pairs(db.groupskills) do
		table.insert(st,k);
	end
	table.sort(st);
	for k,v in pairs(st) do
		if not self.groupSkillOptions[v] then
			self.groupSkillOptions[v] = {
				type = 'group',
				name = v,
				order = k,
				args = self:GetSkillOptions(v, "group");
			};
      self.groupSkillOptions[v].disabled = function() return SorrenTimers.db.char.groups.groupskills.disabled end;
		else
			self.groupSkillOptions[v].order = k;
		end
	end
end

function SorrenTimers:SetEnemySkills()
	local st = {};
	for k,v in pairs(db.customenemy) do
		table.insert(st,k);
	end
	table.sort(st);
	for k,v in pairs(st) do
		if not self.enemySkillOptions[v] then
			self.enemySkillOptions[v] = {
				type = 'group',
				name = v,
				order = k,
				args = self:GetSkillOptions(v, "enemy");
			};
		else
			self.enemySkillOptions[v].order = k;
		end
	end
end

function SorrenTimers:SetBuffs()
	local st = {};
	for k,v in pairs(db.custombuffs) do
		table.insert(st,k);
	end
	table.sort(st);
	for k,v in pairs(st) do
		if not self.buffOptions[v] then
			self.buffOptions[v] = {
				type = 'group',
				name = v,
				order = k,
				args = self:GetSkillOptions(v, "buffs");
			};
		else
			self.buffOptions[v].order = k;
		end
	end
end

function SorrenTimers:GetSkillOptions(k, skillset)
	local skill = db.custom[k];
	local dgroup = db.defaultskillgroup;
	if skillset == "enemy" then
		skill = db.customenemy[k];
		dgroup = db.defaultenemyskillgroup;
	elseif skillset == "buffs" then
		skill = db.custombuffs[k];
		dgroup = db.defaultbuffgroup;
  elseif skillset == "group" then
    skill = db.groupskills[k];
    dgroup = db.defaultgroupskillsgroup;
	end
	local t = {
		enabled = {
			type = 'toggle',
			name = L["Enabled"],
			desc = L["Show a bar for this skill"],
			order = 1,
			get = function()
				return not skill.disabled;
			end,
			set = function(v)
				if v then
					skill.disabled = nil;
				else
					skill.disabled = true;
				end
			end,
		},
		defaultgroup = {
			type = 'toggle',
			name = L["Use Default Group"],
			desc = L["Use the default group for this type of skill"],
			get = function()
				return skill.group == nil;
			end,
			set = function(v)
				if v then
					skill.group = nil;
				else
					skill.group = dgroup;
				end
			end,
			order = 4,
		},
		group = {
			type = 'text',
			order = 5,
			name = L["Group"],
			desc = L["The Group to which this bar belongs"],
			disabled = function() return skill.group == nil end,
			get = function()
				return skill.group or dgroup;
			end,
			set = function(t)
				skill.group = t;
			end,
			validate = self.listGroups,
		},
		defaultcolors = {
			type = 'toggle',
			order = 10,
			name = L["Use Default Colors"],
			desc = L["Use default settings"],
			get = function()
				return skill.colors == nil;
			end,
			set = function(v)
				if not v then
					skill.colors = db.defaults.colors;
				else
					skill.colors = nil;
				end
			end,
		},
		colors = {
			type = 'text',
			order = 11,
			name = L["Colors"],
			desc = L["Colors for this skill\nex.\"green cyan blue\""],
			disabled = function() return skill.colors == nil end,
			get = function()
				local s = "";
				for k,v in pairs(skill.colors or db.defaults.colors) do
					s = s..v.." ";
				end
				return s;
			end,
			set = function(s)
				local v = {};
				for c in string.gmatch(s, "(%w+)") do
					if pc:GetHex(c) then
						table.insert(v,c);
					else
						SorrenTimers:Print(string.format(L["Color %s is not a valid color, skipping."], c));
					end
				end
				if next(v) then
          if skill.colors then del(skill.colors); end
					skill.colors = v;
				else	
					SorrenTimers:Print(L["No colors were valid."]);
				end
			end,
		},
		defaultfade = {
			type = 'toggle',
			order = 15,
			name = L["Use Default Fade Time"],
			desc = L["Use default settings"],
			get = function()
				return skill.fade == nil;
			end,
			set = function(v)
				if not v then
					skill.fade = db.defaults.fade;
				else
					skill.fade = nil;
				end
			end,
		},
		fade = {
			order = 16,
			type = 'text',
			name = L["Fade Time"],
			desc = L["How long it takes the bar to fade after completion, use a negative number if you want it to never fade"],
			disabled = function() return skill.fade == nil; end,
			get = function()
				return skill.fade or db.defaults.fade;
			end,
			set = function(v)
				local n = tonumber(v);
				if not n then 
					SorrenTimers:Print(string.format(L["%s is not a valid number."], v));
				else
					skill.fade = n;
				end
			end,
		},
	};
  if dgroup == db.defaultskillgroup then
    t.share = {
      order = 17,
      type = 'toggle',
      name = L["Share"],
      desc = L["Share this skill duration with others in your group."],
      disabled = function() return self.db.char.groups.groupskills.disabled end,
      get = function()
        return skill.share;
      end,
      set = function(v)
        skill.share = v;
      end,
    };
  end
	return t;
end

function SorrenTimers:OpenOptions()
	db = self.db.char;
	self:SetGroups();
	self:SetSkills();
	self:SetEnemySkills();
	self:SetBuffs();
  self:SetGroupSkills()
  self:SetColors();
	wf:Open('SorrenTimers');
end

function SorrenTimers:SetColors()
  local colors = new();
  for i in ipairs(scolors) do
    scolors[i] = nil;
  end
  for c in pairs(pc.vars.hex) do
    table.insert(colors,c);
  end
  for c in pairs(self.db.char.customcolors) do
    table.insert(scolors,c);
  end
  table.sort(colors);
  table.sort(scolors);
  for i,v in ipairs(colors) do
    if not self.colorOptions[v] then
      self.colorOptions[v] = self:GetColorOptions(v);
    end
    self.colorOptions[v].order = i+10;
  end
  del(colors);
end

function SorrenTimers:GetColorOptions(c)
  local v = {
    type = 'color',
    name = c,
    desc = c,
    get = function()
      local _, r,g,b = pc:GetRGBPercent(c);
      return r,g,b;
    end,
    set = function(r,g,b)
      pc:RegisterColor(c,r,g,b);
      self.db.char.customcolors[c] = new(r,g,b);
    end,
    disabled = function()
      return not self.db.char.customcolors[c];
    end,
  };
  return v;
end

function SorrenTimers:OnTooltipUpdate()
    tablet:SetHint(L["Click for options"]);
end

